/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: hserver.c 89 2003-10-07 15:17:39Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief heart of hserver library
 *  This file contains fork,thread,prefork model handling core.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"
#include "hserver.h"
#include "hslock_sem.h"
#include "hslog.h"
#include "hserver_fork.h"
#include "hserver_thread.h"
#include "hserver_nbio.h"
#include "hserver_prefork.h"

#define HSOCKET(i) ((hsocket*)((i)->data))
#define IS_MODE_FORK(e) ((e)->nType & HS_FORK)
#define IS_MODE_THREAD(e) ((e)->nType & HS_THREAD)
#define IS_MODE_PREFORK(e) ((e)->nType & HS_PREFORK)
#define IS_MODE_NBIO(e) ((e)->nType & HS_NONBLOCKIO)
#define IS_MODE_ASYNCIO(e) ((e)->nType & HS_ASYNCIO)

static void hs_signal(int sig, siginfo_t * sf, void *vp);

/* LOCAL DATA -----------------------------------------------------------------------*/

/*! accept serializing lock */

/*! main hserver engine */
static hserver_engine s_engine;
hserver_engine *g_pEngine = &s_engine;
/*! socket selector */

/* FUNCTION DEFINITION --------------------------------------------------------------*/

hsocket  *hsocket_new(void)
{
	return (hsocket *) malloc(sizeof(hsocket));
}

void hsocket_delete(hsocket * hs)
{
	if (hs)
	{
		free(hs);
	}
	else
	{
		hslog(2, "Null pointer free");
		abort();
	}
}

BOOL hs_init(HSINIT_TYPE coreFunc)
{
	struct sigaction sa;

	hslog(2, "hserver engine init");
	memset(g_pEngine, 0, sizeof *g_pEngine);

	g_pEngine->bTerminated = FALSE;
	g_pEngine->bEventDriven = FALSE;
	hslist_init(&g_pEngine->lstSockets);

	sa.sa_handler = (void (*)(int)) &hs_signal;
	hslog(2, "SIGCHLD/SIGTERM/SIGINT interrupt handler registered");
	sigaction(SIGCHLD, &sa, 0);
	sigaction(SIGTERM, &sa, 0);
	sigaction(SIGINT, &sa, 0);

	/*-- Set default functions */
	g_pEngine->prefork_maintainer = 0;
	g_pEngine->spawn_child = 0;
	g_pEngine->do_task = 0;
	g_pEngine->acceptLock = 0;
	g_pEngine->bEventDriven = FALSE;

	return (*coreFunc) (g_pEngine);
}

BOOL hs_cntl(int nType, int nValue)
{
	switch (nType)
	{
	case HS_PREFORKS:
		g_pEngine->nMaxProcesses = nValue;
		g_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MIN:
		g_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MAX:
		g_pEngine->nMaxProcesses = nValue;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void hs_destroy()
{
	hslist_t  i;

	hslog(2, "hserver shutdown");
	if (!g_pEngine->nSockets)
	{
		return;
	}
	HSLIST_FOR_EACH(i, g_pEngine->lstSockets)
	{
		if (HSOCKET(i)->fd != (SOCKET) - 1)
		{
			if (close(HSOCKET(i)->fd))
			{
				hslog(1, "server socket [%d] closing error", HSOCKET(i)->fd);
			}
		}
	}
}

void hs_task_install(hsocket * s, hs_callback pfCallBack)
{
	if (!s)
	{
		return;
	}
	s->callback = pfCallBack;
	s->vft = g_pEngine->sock_vft;
	hslist_push_front(&g_pEngine->lstSockets, s);
	g_pEngine->nSockets++;
}

hsocket  *hsocket_server(const char *ip, int port)
{
	hsocket  *sock;
	int       reuse = 1;
	struct sockaddr_in addr;

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	inet_aton(ip, &(addr.sin_addr));

	sock = hsocket_new();
	if (!sock)
	{
		return 0;
	}

	sock->fd = socket(PF_INET, SOCK_STREAM, 0);
	if (sock->fd < 0)
	{
		hserrlog("socket");
	}
	else if (0 > setsockopt(sock->fd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
	}
	else if (0 > bind(sock->fd, (struct sockaddr *) &addr, sizeof addr))
	{
		hserrlog("bind to (%s:%d)", ip, port);
		free(sock);
		return 0;
	}
	sock->state = HSS_LISTEN;
	return sock;
}

static BOOL listen_sockets()
{
	hslist_t  i;

	if (HSLIST_EMPTY(g_pEngine->lstSockets))
	{
		return FALSE;
	}
	HSLIST_FOR_EACH(i, g_pEngine->lstSockets)
	{
		assert(HSOCKET(i)->fd >= 0);
		if (HSOCKET(i)->state == HSS_LISTEN && 0 > listen(HSOCKET(i)->fd, 5))
		{
			return FALSE;
		}
		hslog(2, "listen %d", HSOCKET(i)->fd);
	}
	return TRUE;
}

static BOOL select_sockets()
{
	hslist_t  i;
	int       max = -1;
	int       nCount = 0;
	fd_set    fdRead;

      again:
	FD_ZERO(&fdRead);

	hslog(1, "select sockets: %d", g_pEngine->nSockets);

	/* Prefork mode, listening 2 or more sockets */
	if (IS_MODE_PREFORK(g_pEngine) && g_pEngine->nSockets > 1)
	{
		if (!g_pEngine->lock_vft->lock(g_pEngine->acceptLock))
		{
			hslog(1, "Can't serializing lock");
			return FALSE;
		}
	}

	HSLIST_FOR_EACH(i, g_pEngine->lstSockets)
	{
		SOCKET    s = HSOCKET(i)->fd;
		assert(s);
		FD_SET(s, &fdRead);
		if (max < s)
		{
			max = s;
		}
		HSOCKET(i)->interesting = FALSE;
	}

	if (0 >= (nCount = select(max + 1, &fdRead, 0, 0, 0)))
	{
		if (errno == EINTR && !g_pEngine->bTerminated)
		{
			goto again;
		}
		hserrlog("Select error");
		return FALSE;
	}
	hslog(2, "selected socket count [%d]", nCount);

	if (IS_MODE_PREFORK(g_pEngine) && g_pEngine->nSockets > 1)
	{
		if (!g_pEngine->lock_vft->unlock(g_pEngine->acceptLock))
		{
			hslog(1, "Can't serializing unlock");
			return FALSE;
		}
	}

	g_pEngine->nInterestings = nCount;
	HSLIST_FOR_EACH(i, g_pEngine->lstSockets)
	{
		hs_vfunc  sockhandler;
		if (!FD_ISSET(HSOCKET(i)->fd, &fdRead))
		{
			continue;
		}
		if (!HSOCKET(i)->vft)
		{
			hslog(1, "Error no vft after event.");
			abort();
		}
		sockhandler = (*(HSOCKET(i)->vft))[(int) HSHANDLE_ONRECV];
		if (sockhandler)
		{
			hslog(2, "Handler invoking..");
			if (!(*sockhandler) (HSOCKET(i)))
			{
				break;	/* HSLIST_FOR_EACH */
			}
		}
		else
		{
			hslog(1, "No function table");
			abort();
		}
		HSOCKET(i)->interesting = TRUE;
		--nCount;
		if (nCount <= 0)
		{
			break;	/* HSLIST_FOR_EACH */
		}
	}

	return TRUE;
}

static void hs_signal(int sig, siginfo_t * sf, void *vp)
{
	pid_t     pid;
	int       status;

	sf = sf;
	vp = vp;

	switch (sig)
	{
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		{
			hslog(2, "worker terminated [pid:%d]", pid);
			g_pEngine->nCurrProcesses--;

			/* none */
		}
		break;
	case SIGTERM:
	case SIGINT:
		g_pEngine->bTerminated = TRUE;
		hs_cntl(HS_PREFORKS, 0);
		hslog(2, "Terminating..");
		break;
	default:
		hserrlog("Unknown signal caught [%d]", sig);
		assert(0 && "Can't handle this signal");
		break;
	}
}

static void hs_poll()
{
	/* Main loop */
	while (!g_pEngine->bTerminated)
	{
		if (!select_sockets())
		{
			return;
		}
		if (g_pEngine->do_task)
		{
/*			g_pEngine->do_task( g_pEngine );*/
		}
	}
}

static void hs_wait_event()
{
	while (1)
		sleep(10);
}

/*! \brief main exported run function */
void hs_run()
{
	if (FALSE == listen_sockets())
	{
		return;
	}

	if (IS_MODE_PREFORK(g_pEngine) && g_pEngine->prefork_maintainer)
	{
		if (g_pEngine->nSockets > 1)
		{
			g_pEngine->acceptLock = g_pEngine->lock_vft->init((int *) &g_pEngine->bTerminated);
			if (FALSE == g_pEngine->acceptLock)
			{
				hslog(2, "Lock initialization failure.");
				return;
			}
		}

		if (g_pEngine->prefork_maintainer(g_pEngine))
		{
			/* pre-forked main server return */
			goto final;
		}
		/* pre-forked child passes thru.. */
	}

	hs_poll();

	if (g_pEngine->bEventDriven)
	{
		hs_wait_event();
	}

      final:
	if (g_pEngine->acceptLock)
	{
		if (FALSE == g_pEngine->lock_vft->destroy(g_pEngine->acceptLock))
		{
			hslog(2, "Lock destroying failure");
		}
		g_pEngine->acceptLock = NULL;
	}

}

/*
 * History
   $Log$
   Revision 1.3  2003/10/07 15:17:39  pynoos
   Bug fix in freeing lock

   Revision 1.2  2003/10/07 08:51:02  pynoos
   First draft of BLOCKED-TYPE series

 */

/*
 * vim: set path+=../src
 *      set tags+=../tags
 */
