/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: model2.c 51 2003-09-17 13:46:21Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief Model 2 thinkpad..
 *  model2 is accept-thread-creation.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"
#include "hslog.h"

typedef int hs_callback(int fd, void *arg);

typedef struct _hs_thread_pv
{
	pthread_t thrID;
	int       sockWorker;
	void     *pArg;
	hs_callback *pfWorker;
}
hs_thread_pv;

typedef struct _hserver_bind
{
	int       sock;
	BOOL      interesting;
	hs_callback *worker;
	void     *arg;
	struct sockaddr_in address;
}
hserver_bind;

typedef struct _hserver_engine
{
	int       nBinds;
	int       nInterestings;
	hserver_bind *pBind;
}
hserver_engine_t;

/* LOCAL DATA -----------------------------------------------------------------------*/

static hserver_engine_t s_engine;

/* FUNCTION DEFINITION --------------------------------------------------------------*/

static void hs_signal(int sig, siginfo_t * sf, void *vp)
{
	pid_t     pid;
	int       status;

	switch (sig)
	{
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		{
			hslog(2, "worker terminated [pid:%d]", pid);
			/* none */
		}
		break;
	default:
		hserrlog("Unknown signal caught [%d:%s]", sig, strsignal(sig));
		assert(0 && "Can't handle this signal");
		break;
	}
}

void hs_init(hserver_engine_t * pEngine)
{
	struct sigaction sa;

	assert(pEngine);

	hslog(2, "hserver engine init");
	memset(pEngine, 0, sizeof *pEngine);

	sa.sa_sigaction = &hs_signal;
	hslog(2, "SIGCHLD interrupt handler registered");
	sigaction(SIGCHLD, &sa, 0);
}

void hs_destroy(hserver_engine_t * pEngine)
{
	int       i;

	hslog(2, "hserver shutdown");
	if (!pEngine->pBind)
	{
		return;
	}
	for (i = 0; i < pEngine->nBinds; i++)
	{
		if (pEngine->pBind[i].sock != (SOCKET) - 1)
		{
			if (close(pEngine->pBind[i].sock))
			{
				hslog(1, "server socket [%d] closing error", pEngine->pBind[i].sock);
			}
		}
	}
	free(pEngine->pBind);
}

void hs_server_add(hserver_engine_t * pEngine, const char *ip, int nPort, hs_callback * pfWorker, void *pArg)
{
	hserver_bind *pNewBinds;

	assert(pfWorker);
	assert(pEngine);

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	hslog(2, "server socket added [%s:%d]", ip, nPort);

	pNewBinds = (hserver_bind *) malloc(sizeof(hserver_bind) * (pEngine->nBinds + 1));
	if (pEngine->pBind)
	{
		memcpy(pNewBinds, pEngine->pBind, sizeof(hserver_bind) * pEngine->nBinds);
		free(pEngine->pBind);
#ifndef NDEBUG
		pEngine->pBind = 0;
#endif
	}
	pEngine->pBind = pNewBinds;
	pEngine->pBind[pEngine->nBinds].address.sin_family = AF_INET;
	pEngine->pBind[pEngine->nBinds].address.sin_port = htons(nPort);
	pEngine->pBind[pEngine->nBinds].worker = pfWorker;
	pEngine->pBind[pEngine->nBinds].arg = pArg;
	inet_aton(ip, &(pEngine->pBind[pEngine->nBinds].address.sin_addr));
	pEngine->nBinds++;
}

static SOCKET create_a_socket(const struct sockaddr_in *pAddr)
{
	SOCKET    s;
	int       reuse = 1;

	assert(pAddr);

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		hserrlog("socket");
		return s;
	}
	if (0 > setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
		return -1;
	}
	if (0 > bind(s, (struct sockaddr *) pAddr, sizeof *pAddr))
	{
		hserrlog("bind");
		return -1;
	}
	return s;
}

static BOOL create_sockets(hserver_engine_t * pEngine)
{
	int       i;

	assert(pEngine);

	if (0 >= pEngine->pBind)
	{
		TODO("Error..");
		return FALSE;
	}

	for (i = 0; i < pEngine->nBinds; i++)
	{
		pEngine->pBind[i].sock = create_a_socket(&(pEngine->pBind[i].address));
	}
	return TRUE;
}

static BOOL listen_sockets(hserver_engine_t * pEngine)
{
	int       i;
	assert(pEngine);

	for (i = 0; i < pEngine->nBinds; i++)
	{
		assert(pEngine->pBind[i].sock >= 0);
		if (0 > listen(pEngine->pBind[i].sock, 5))
		{
			return FALSE;
		}
		hslog(2, "listen %d", i + 1);
	}
	return TRUE;
}

static BOOL select_sockets(hserver_engine_t * pEngine)
{
	int       i;
	int       max = -1;
	int       nCount = 0;
	fd_set    fdRead;

	assert(pEngine);
      again:
	FD_ZERO(&fdRead);

	hslog(1, "select sockets: %d", pEngine->nBinds);
	for (i = 0; i < pEngine->nBinds; i++)
	{
		SOCKET    s = pEngine->pBind[i].sock;
		assert(s);
		FD_SET(s, &fdRead);
		if (max < s)
		{
			max = s;
		}
		pEngine->pBind[i].interesting = FALSE;
	}

	if (0 >= (nCount = select(max + 1, &fdRead, 0, 0, 0)))
	{
		if (errno == EINTR)
		{
			goto again;
		}
		return FALSE;
	}
	hslog(2, "selected socket count [%d]", nCount);

	pEngine->nInterestings = nCount;
	for (i = 0; i < pEngine->nBinds && nCount > 0; i++)
	{
		if (FD_ISSET(pEngine->pBind[i].sock, &fdRead))
		{
			pEngine->pBind[i].interesting = TRUE;
			nCount--;
		}
	}

	return TRUE;
}

static BOOL select_a_socket(SOCKET sock, int timeout)
{
	fd_set    fdRead, fdWrite, fdExcpt;
	struct timeval tval;
	int       ret;

	FD_ZERO(&fdRead);
	FD_ZERO(&fdWrite);
	FD_ZERO(&fdExcpt);

	FD_SET(sock, &fdRead);
	FD_SET(sock, &fdWrite);
	FD_SET(sock, &fdExcpt);

	if (timeout >= 0)
	{
		tval.tv_sec = timeout;
		tval.tv_usec = 0;
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, &tval);
	}
	else
	{
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, 0);
	}
	return ret == 1;
}

static void *thread_entrance(void *pv)
{
	hs_thread_pv *pvThreadArg = (hs_thread_pv *) pv;
	hs_callback *pfWorker = pvThreadArg->pfWorker;
	SOCKET    sockWorker = pvThreadArg->sockWorker;
	void     *pArg = pvThreadArg->pArg;

	pthread_detach(pvThreadArg->thrID);
	(*pfWorker) (sockWorker, pArg);
	sleep(0);
	shutdown(sockWorker, 2);
	hslog(2, "wait peer close");
	select_a_socket(sockWorker, 10);
	close(sockWorker);
	hslog(2, "thread exit [threadid:%d]", pvThreadArg->thrID);
	pthread_exit(0);
}

static BOOL spawn_child_t(hserver_engine_t * pEngine, int index, SOCKET sockWorker)
{
	int       i;
	hs_thread_pv *pvThreadArg;
	BOOL      bSuccessSpawn = FALSE;
	pthread_attr_t attrThread;

	assert(pEngine);
	assert(index >= 0);

	pvThreadArg = (hs_thread_pv *) malloc(sizeof *pvThreadArg);
	if (!pvThreadArg)
	{
		hslog(1, "Not enough memory");
		return FALSE;
	}

	assert(sockWorker != (SOCKET) - 1);
	pvThreadArg->sockWorker = sockWorker;
	pvThreadArg->pArg = pEngine->pBind[index].arg;
	pvThreadArg->pfWorker = pEngine->pBind[index].worker;

	pthread_attr_init(&attrThread);
	pthread_attr_setdetachstate(&attrThread, PTHREAD_CREATE_DETACHED);

	/* Try to fork a process */
	while (!bSuccessSpawn)
	{
		switch (pthread_create(&pvThreadArg->thrID, &attrThread, thread_entrance, pvThreadArg))
		{
		case 0:
			hslog(2, "thread created [threadid:%d]", pvThreadArg->thrID);
			bSuccessSpawn = TRUE;
			break;
		default:
			hserrlog("pthread_create");
			sleep(1);
			continue;
		}
	}

}

static BOOL do_service(hserver_engine_t * pEngine)
{
	int       n;
	int       i;

	assert(pEngine);

	n = pEngine->nInterestings;

	if (n <= 0)
	{
		return FALSE;
	}

	for (i = 0; i < pEngine->nBinds && n > 0; i++)
	{
		if (!pEngine->pBind[i].interesting)
		{
			continue;
		}

		n--;

		/* Selectable interface */
		{
			SOCKET    ws;
			struct sockaddr_in addr;
			int       size = sizeof addr;

			ws = accept(pEngine->pBind[i].sock, (struct sockaddr *) &addr, &size);
			spawn_child_t(pEngine, i, ws);
		}
		pEngine->pBind[i].interesting = FALSE;
	}
}

void hs_run(hserver_engine_t * pEngine)
{
	if (FALSE == create_sockets(pEngine))
	{
		return;
	}
	if (FALSE == listen_sockets(pEngine))
	{
		return;
	}
	while (1)
	{
		if (!select_sockets(pEngine))
		{
			continue;
		}
		do_service(pEngine);
	}
}

/*------------- USER ---------------*/

int webserver(int fd, void *arg)
{
	const char *msg = "200 OK\r\nContent-type: text/plain\r\n\r\nTest page\r\n";
	char      req[LARGE_BUFFER_SIZE];
	if (0 < recv(fd, req, sizeof req, 0))
	{
		send(fd, msg, strlen(msg), 0);
	}
	return 0;
}

int main()
{
	hserver_engine_t engine;

	hs_init(&engine);
	hs_server_add(&engine, "0.0.0.0", 8080, webserver, (void *) 0);
	hs_server_add(&engine, "0.0.0.0", 8088, webserver, (void *) 0);
	hs_run(&engine);
	hs_destroy(&engine);
}

/*
 * vim: set path+=../src
 */
