/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: model3.c 54 2003-09-22 16:32:55Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief Model 3 thinkpad..
 *  model3 is prefork-accept, another historical model.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"
#include "hslog.h"

typedef int hs_callback(int fd, void *arg);

typedef struct _hserver_bind
{
	int       sock;
	BOOL      interesting;
	hs_callback *worker;
	void     *arg;
	struct sockaddr_in address;
}
hserver_bind;

typedef struct _hserver_engine
{
	int       nType;
	int       nBinds;
	int       nInterestings;
	int       nMaxProcesses;
	int       nMinProcesses;
	int       nCurrProcesses;
	hserver_bind *pBind;
	BOOL      bTerminated;
}
hserver_engine_t;

enum _hs_type
{
	HS_FORK,
	HS_THREAD,
	HS_PREFORKED
};

enum _hs_cntl_type
{
	HS_PREFORKS,
	HS_PREFORK_MIN,
	HS_PREFORK_MAX
};

/* LOCAL DATA -----------------------------------------------------------------------*/

static hserver_engine_t s_engine;
static hserver_engine_t *s_pEngine = &s_engine;

/* FUNCTION DEFINITION --------------------------------------------------------------*/

static void hs_signal(int sig, siginfo_t * sf, void *vp)
{
	pid_t     pid;
	int       status;

	switch (sig)
	{
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		{
			hslog(2, "worker terminated [pid:%d]", pid);
			s_pEngine->nCurrProcesses--;

			/* none */
		}
		break;
	case SIGTERM:
	case SIGINT:
		s_pEngine->bTerminated = TRUE;
		hs_cntl(HS_PREFORKS, 0);
		hslog(2, "Terminating..");
		break;
	default:
		hserrlog("Unknown signal caught [%d:%s]", sig, strsignal(sig));
		assert(0 && "Can't handle this signal");
		break;
	}
}

BOOL hs_init(int nType)
{
	struct sigaction sa;

	assert(s_pEngine);

	hslog(2, "hserver engine init");
	memset(s_pEngine, 0, sizeof *s_pEngine);

	s_pEngine->nType = nType;
	s_pEngine->bTerminated = FALSE;

	sa.sa_sigaction = &hs_signal;
	hslog(2, "SIGCHLD/SIGTERM/SIGINT interrupt handler registered");
	sigaction(SIGCHLD, &sa, 0);
	sigaction(SIGTERM, &sa, 0);
	sigaction(SIGINT, &sa, 0);
}

BOOL hs_cntl(int nType, int nValue)
{
	assert(s_pEngine);

	switch (nType)
	{
	case HS_PREFORKS:
		s_pEngine->nMaxProcesses = nValue;
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MIN:
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MAX:
		s_pEngine->nMaxProcesses = nValue;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void hs_destroy()
{
	int       i;

	hslog(2, "hserver shutdown");
	if (!s_pEngine->pBind)
	{
		return;
	}
	for (i = 0; i < s_pEngine->nBinds; i++)
	{
		if (s_pEngine->pBind[i].sock != (SOCKET) - 1)
		{
			if (close(s_pEngine->pBind[i].sock))
			{
				hslog(1, "server socket [%d] closing error", s_pEngine->pBind[i].sock);
			}
		}
	}
	free(s_pEngine->pBind);
}

void hs_server_add(const char *ip, int nPort, hs_callback * pfWorker, void *pArg)
{
	hserver_bind *pNewBinds;

	assert(pfWorker);
	assert(s_pEngine);

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	hslog(2, "server socket added [%s:%d]", ip, nPort);

	pNewBinds = (hserver_bind *) malloc(sizeof(hserver_bind) * (s_pEngine->nBinds + 1));
	if (s_pEngine->pBind)
	{
		memcpy(pNewBinds, s_pEngine->pBind, sizeof(hserver_bind) * s_pEngine->nBinds);
		free(s_pEngine->pBind);
#ifndef NDEBUG
		s_pEngine->pBind = 0;
#endif
	}
	s_pEngine->pBind = pNewBinds;
	s_pEngine->pBind[s_pEngine->nBinds].address.sin_family = AF_INET;
	s_pEngine->pBind[s_pEngine->nBinds].address.sin_port = htons(nPort);
	s_pEngine->pBind[s_pEngine->nBinds].worker = pfWorker;
	s_pEngine->pBind[s_pEngine->nBinds].arg = pArg;
	inet_aton(ip, &(s_pEngine->pBind[s_pEngine->nBinds].address.sin_addr));
	s_pEngine->nBinds++;
}

static SOCKET create_a_socket(const struct sockaddr_in *pAddr)
{
	SOCKET    s;
	int       reuse = 1;

	assert(pAddr);

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		hserrlog("socket");
		return s;
	}
	if (0 > setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
		return -1;
	}
	if (0 > bind(s, (struct sockaddr *) pAddr, sizeof *pAddr))
	{
		hserrlog("bind");
		return -1;
	}
	return s;
}

static BOOL create_sockets()
{
	int       i;

	assert(s_pEngine);

	if (0 >= s_pEngine->pBind)
	{
		TODO("Error..");
		return FALSE;
	}

	for (i = 0; i < s_pEngine->nBinds; i++)
	{
		if (0 > (s_pEngine->pBind[i].sock = create_a_socket(&(s_pEngine->pBind[i].address))))
		{
			return FALSE;
		}
	}
	return TRUE;
}

static BOOL listen_sockets()
{
	int       i;
	assert(s_pEngine);

	for (i = 0; i < s_pEngine->nBinds; i++)
	{
		assert(s_pEngine->pBind[i].sock >= 0);
		if (0 > listen(s_pEngine->pBind[i].sock, 5))
		{
			return FALSE;
		}
		hslog(2, "listen %d", i + 1);
	}
	return TRUE;
}

static int hs_lock;

static void hs_lock_close(void)
{
	semctl(hs_lock, 0, IPC_RMID, 0);
}

static void hs_lock_init()
{
	union semun semval;
	hs_lock = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
	if (hs_lock < 0)
	{
		hserrlog("semget");
		return;
	}
	semval.val = 1;
	if (semctl(hs_lock, 0, SETVAL, semval) < 0)
	{
		hserrlog("semctl");
		return;
	}
	atexit(hs_lock_close);
}

static void hs_lock_hold()
{
	static struct sembuf ops = { 0, -1, SEM_UNDO };
	while (0 > semop(hs_lock, &ops, 1))
	{
		if (errno == EINTR)
		{
			continue;
		}
		hserrlog("semop");
		return;
	}
}

static void hs_lock_release()
{
	static struct sembuf ops = { 0, 1, SEM_UNDO };
	while (0 > semop(hs_lock, &ops, 1))
	{
		if (errno == EINTR)
		{
			continue;
		}
		hserrlog("semop");
		return;
	}
}

static BOOL select_sockets()
{
	int       i;
	int       max = -1;
	int       nCount = 0;
	fd_set    fdRead;

	assert(s_pEngine);
      again:
	FD_ZERO(&fdRead);

	hslog(1, "select sockets: %d", s_pEngine->nBinds);

	if (s_pEngine->nBinds > 1)
	{
		hs_lock_hold();
	}

	for (i = 0; i < s_pEngine->nBinds; i++)
	{
		SOCKET    s = s_pEngine->pBind[i].sock;
		assert(s);
		FD_SET(s, &fdRead);
		if (max < s)
		{
			max = s;
		}
		s_pEngine->pBind[i].interesting = FALSE;
	}

	if (0 >= (nCount = select(max + 1, &fdRead, 0, 0, 0)))
	{
		if (errno == EINTR && !s_pEngine->bTerminated)
		{
			goto again;
		}
		return FALSE;
	}
	hslog(2, "selected socket count [%d]", nCount);

	if (s_pEngine->nBinds > 1)
	{
		hs_lock_release();
	}

	s_pEngine->nInterestings = nCount;
	for (i = 0; i < s_pEngine->nBinds && nCount > 0; i++)
	{
		if (FD_ISSET(s_pEngine->pBind[i].sock, &fdRead))
		{
			s_pEngine->pBind[i].interesting = TRUE;
			nCount--;
		}
	}

	return TRUE;
}

static BOOL select_a_socket(SOCKET sock, int timeout)
{
	fd_set    fdRead, fdWrite, fdExcpt;
	struct timeval tval;
	int       ret;

	FD_ZERO(&fdRead);
	FD_ZERO(&fdWrite);
	FD_ZERO(&fdExcpt);

	FD_SET(sock, &fdRead);
	FD_SET(sock, &fdWrite);
	FD_SET(sock, &fdExcpt);

	if (timeout >= 0)
	{
		tval.tv_sec = timeout;
		tval.tv_usec = 0;
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, &tval);
	}
	else
	{
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, 0);
	}
	return ret == 1;
}

static BOOL spawn_child(int index, SOCKET sockWorker)
{
	int       i;
	hs_callback *pfWorker;
	void     *pArg;
	BOOL      bSuccessFork = FALSE;
	pid_t     pid;

	assert(s_pEngine);
	assert(index >= 0);

	/* Try to fork a process */
	while (!bSuccessFork)
	{
		switch ((pid = fork()))
		{
		case -1:
			hserrlog("fork");
			sleep(1);
			continue;
		case 0:
			bSuccessFork = TRUE;
			break;
		default:
			hslog(2, "worker forked [pid:%d]", pid);
			close(sockWorker);
			return TRUE;
		}
	}
	/* end fork; child process will go on */

}

static BOOL do_service_forked()
{
	SOCKET    sockWorker = (SOCKET) - 1;
	hs_callback *pfWorker;
	void     *pArg;

	int       n;
	int       i;

	assert(s_pEngine);

	n = s_pEngine->nInterestings;

	if (n <= 0)
	{
		return FALSE;
	}

	for (i = 0; i < s_pEngine->nBinds && n > 0; i++)
	{
		if (!s_pEngine->pBind[i].interesting)
		{
			continue;
		}

		/* Selectable interface */
		{
			struct sockaddr_in addr;
			int       size = sizeof addr;

			sockWorker = accept(s_pEngine->pBind[i].sock, (struct sockaddr *) &addr, &size);
			break;
		}
		n--;
		s_pEngine->pBind[i].interesting = FALSE;
	}

	assert(i < s_pEngine->nBinds && n > 0);

	pfWorker = s_pEngine->pBind[i].worker;
	pArg = s_pEngine->pBind[i].arg;

	assert(sockWorker != (SOCKET) - 1);
	assert(pfWorker);

	(*pfWorker) (sockWorker, pArg);
	sleep(0);
	shutdown(sockWorker, 2);
	hslog(2, "wait peer close");
	select_a_socket(sockWorker, 10);
	close(sockWorker);
}

void hs_run_forked()
{
	while (!s_pEngine->bTerminated)
	{
		if (!select_sockets(s_pEngine))
		{
			continue;
		}
		do_service_forked(s_pEngine);
	}
}

static void keep_processes()
{
	while (!s_pEngine->bTerminated && s_pEngine->nCurrProcesses < s_pEngine->nMinProcesses)
	{
		pid_t     pid;
		switch ((pid = fork()))
		{
		case -1:
			hserrlog("fork");
			sleep(1);
			continue;
		case 0:
			hs_run_forked(s_pEngine);
			exit(0);
		default:
			s_pEngine->nCurrProcesses++;
			hslog(2, "forked [pid:%d]", pid);
		}
	}
}

void hs_run()
{
	if (FALSE == create_sockets(s_pEngine))
	{
		return;
	}
	if (FALSE == listen_sockets(s_pEngine))
	{
		return;
	}

	if (s_pEngine->nBinds > 1)
	{
		hs_lock_init();
	}
	while (!s_pEngine->bTerminated)
	{
		keep_processes(s_pEngine);
		sleep(300);
	}
}

/*------------- USER ---------------*/

int webserver(int fd, void *arg)
{
	const char *msg = "200 OK\r\nContent-type: text/plain\r\n\r\nTest page\r\n";
	char      req[LARGE_BUFFER_SIZE];
	if (0 < recv(fd, req, sizeof req, 0))
	{
		send(fd, msg, strlen(msg), 0);
	}
	return 0;
}

int main()
{
	hs_init(HS_PREFORKED);
	hs_cntl(HS_PREFORKS, 5);
	hs_server_add("0.0.0.0", 8080, webserver, (void *) 0);
	hs_server_add("0.0.0.0", 8088, webserver, (void *) 0);
	hs_run();
	hs_destroy();
}

/*
 * vim: set path+=../src
 */
