/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: model31.c 62 2003-09-24 16:44:26Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief Model 4 thinkpad..
 *  model4 is one process non-block I/O type.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"
#include "hslog.h"
#include "hslist.h"

enum _HS_STATE
{
	HSS_STATE_BEGIN = 1000,
	HSS_INIT,
	HSS_ONACCEPT,
	HSS_RECV,
	HSS_ONRECV,
	HSS_SEND,
	HSS_ONSENT,
	HSS_CLOSE,
	HSS_ONCLOSE,
	HSS_ONERROR,
	HSS_STATE_END
};
typedef enum _HS_STATE hs_state_t;

typedef struct _hs_ctx_t hs_ctx_t;
typedef hs_state_t hs_handler(hs_ctx_t *);

struct _hs_ctx_t
{
	hs_state_t state;				   /* CURRENT STATE */
	int       sock;					   /* SOCKET */
	BOOL      interesting;				   /* SELECTED FOR ONACCEPT, ONRECV, ONCLOSE */
	int       errnum;				   /* ONERROR */
	struct sockaddr_in address;			   /* SERVER SOCKET BIND ADDRESS */

	          hs_state_t(*handler) (hs_ctx_t *);	   /* SOCKET HANDLER */
	void     *argp;					   /* FOR USER USAGE */
	long      argl;					   /* FOR USER USAGE */
};

typedef struct _hserver_engine
{
	int       nType;
	int       nInterestings;
	int       nMaxProcesses;
	int       nMinProcesses;
	int       nCurrProcesses;
	int       nSockets;
	hslist_t  lstSockets;
	BOOL      bTerminated;
}
hserver_engine_t;

#define HSCTX(i) ((hs_ctx_t*)((i)->data))

enum _hs_type
{
	HS_FORK,
	HS_THREAD,
	HS_PREFORKED,
	HS_NONBLOCKIO,
	HS_PREFORKED_NONBLOCKIO,
	HS_THREAD_NONBLOCKIO
};

enum _hs_cntl_type
{
	HS_PREFORKS,
	HS_PREFORK_MIN,
	HS_PREFORK_MAX
};

/* LOCAL DATA -----------------------------------------------------------------------*/

static hserver_engine_t s_engine;
static hserver_engine_t *s_pEngine = &s_engine;

/* FUNCTION DEFINITION --------------------------------------------------------------*/

hs_ctx_t *hs_ctx_new(void)
{
	return (hs_ctx_t *) malloc(sizeof(struct _hs_ctx_t));
}

static void hs_signal(int sig, siginfo_t * sf, void *vp)
{
	pid_t     pid;
	int       status;

	switch (sig)
	{
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		{
			hslog(2, "worker terminated [pid:%d]", pid);
			s_pEngine->nCurrProcesses--;

			/* none */
		}
		break;
	case SIGTERM:
	case SIGINT:
		s_pEngine->bTerminated = TRUE;
		hs_cntl(HS_PREFORKS, 0);
		hslog(2, "Terminating..");
		break;
	default:
		hserrlog("Unknown signal caught [%d:%s]", sig, strsignal(sig));
		assert(0 && "Can't handle this signal");
		break;
	}
}

BOOL hs_init(int nType)
{
	struct sigaction sa;

	assert(s_pEngine);

	hslog(2, "hserver engine init");
	memset(s_pEngine, 0, sizeof *s_pEngine);

	s_pEngine->nType = nType;
	s_pEngine->bTerminated = FALSE;
	hslist_init(&s_pEngine->lstSockets);

	sa.sa_sigaction = &hs_signal;
	hslog(2, "SIGCHLD/SIGTERM/SIGINT interrupt handler registered");
	sigaction(SIGCHLD, &sa, 0);
	sigaction(SIGTERM, &sa, 0);
	sigaction(SIGINT, &sa, 0);
}

BOOL hs_cntl(int nType, int nValue)
{
	assert(s_pEngine);

	switch (nType)
	{
	case HS_PREFORKS:
		s_pEngine->nMaxProcesses = nValue;
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MIN:
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MAX:
		s_pEngine->nMaxProcesses = nValue;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void hs_destroy()
{
	hslist_t  i;

	hslog(2, "hserver shutdown");
	if (!s_pEngine->nSockets)
	{
		return;
	}
	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (HSCTX(i)->sock != (SOCKET) - 1)
		{
			if (close(HSCTX(i)->sock))
			{
				hslog(1, "server socket [%d] closing error", HSCTX(i)->sock);
			}
		}
	}
}

void hs_server_add(const char *ip, int nPort, hs_handler * pfHandler)
{
	hs_ctx_t *ctx;

	assert(pfHandler);

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	hslog(2, "server socket added [%s:%d]", ip, nPort);

	ctx = hs_ctx_new();
	if (!ctx)
	{
		TODO("Memory allocation failure.");
		return;
	}
	ctx->address.sin_family = AF_INET;
	ctx->address.sin_port = htons(nPort);
	ctx->handler = pfHandler;
	inet_aton(ip, &(ctx->address.sin_addr));
	hslist_push_front(&s_pEngine->lstSockets, ctx);
	s_pEngine->nSockets++;
}

static SOCKET create_a_socket(const struct sockaddr_in *pAddr)
{
	SOCKET    s;
	int       reuse = 1;

	assert(pAddr);

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		hserrlog("socket");
		return s;
	}
	if (0 > setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
		return -1;
	}
	if (0 > bind(s, (struct sockaddr *) pAddr, sizeof *pAddr))
	{
		hserrlog("bind");
		return -1;
	}
	return s;
}

static BOOL create_sockets()
{
	hslist_t  i;

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (0 > (HSCTX(i)->sock = create_a_socket(&(HSCTX(i)->address))))
		{
			return FALSE;
		}
	}
	return TRUE;
}

static BOOL listen_sockets()
{
	hslist_t  i;

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		assert(HSCTX(i)->sock >= 0);
		if (0 > listen(HSCTX(i)->sock, 5))
		{
			return FALSE;
		}
		hslog(2, "listen %d", HSCTX(i)->sock);
	}
	return TRUE;
}

static int hs_lock;

static void hs_lock_close(void)
{
	semctl(hs_lock, 0, IPC_RMID, 0);
}

static void hs_lock_init()
{
	union semun semval;
	hs_lock = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
	if (hs_lock < 0)
	{
		hserrlog("semget");
		return;
	}
	semval.val = 1;
	if (semctl(hs_lock, 0, SETVAL, semval) < 0)
	{
		hserrlog("semctl");
		return;
	}
	atexit(hs_lock_close);
}

static BOOL hs_lock_hold()
{
	static struct sembuf ops = { 0, -1, SEM_UNDO };
	while (0 > semop(hs_lock, &ops, 1))
	{
		if (errno == EINTR)
		{
			continue;
		}
		hserrlog("semop");
		return FALSE;
	}
	return TRUE;
}

static BOOL hs_lock_release()
{
	static struct sembuf ops = { 0, 1, SEM_UNDO };
	while (0 > semop(hs_lock, &ops, 1))
	{
		if (errno == EINTR)
		{
			continue;
		}
		hserrlog("semop");
		return FALSE;
	}
	return TRUE;
}

static BOOL select_sockets()
{
	hslist_t  i;
	int       max = -1;
	int       nCount = 0;
	fd_set    fdRead;

	assert(s_pEngine);
      again:
	FD_ZERO(&fdRead);

	hslog(1, "select sockets: %d", s_pEngine->nSockets);

	if (s_pEngine->nSockets > 1)
	{
		if (!hs_lock_hold())
		{
			return FALSE;
		}
	}

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		SOCKET    s = HSCTX(i)->sock;
		assert(s);
		FD_SET(s, &fdRead);
		if (max < s)
		{
			max = s;
		}
		HSCTX(i)->interesting = FALSE;
	}

	if (0 >= (nCount = select(max + 1, &fdRead, 0, 0, 0)))
	{
		if (errno == EINTR && !s_pEngine->bTerminated)
		{
			goto again;
		}
		return FALSE;
	}
	hslog(2, "selected socket count [%d]", nCount);

	if (s_pEngine->nSockets > 1)
	{
		if (!hs_lock_release())
		{
			return FALSE;
		}
	}

	s_pEngine->nInterestings = nCount;
	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (FD_ISSET(HSCTX(i)->sock, &fdRead))
		{
			HSCTX(i)->interesting = TRUE;
			nCount--;
		}
	}

	return TRUE;
}

static BOOL select_a_socket(SOCKET sock, int timeout)
{
	fd_set    fdRead, fdWrite, fdExcpt;
	struct timeval tval;
	int       ret;

	FD_ZERO(&fdRead);
	FD_ZERO(&fdWrite);
	FD_ZERO(&fdExcpt);

	FD_SET(sock, &fdRead);
	FD_SET(sock, &fdWrite);
	FD_SET(sock, &fdExcpt);

	if (timeout >= 0)
	{
		tval.tv_sec = timeout;
		tval.tv_usec = 0;
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, &tval);
	}
	else
	{
		ret = select(sock + 1, &fdRead, &fdWrite, &fdExcpt, 0);
	}
	return ret == 1;
}

static BOOL spawn_child(int index, SOCKET sockWorker)
{
	int       i;
	hs_handler *pfHandler;
	BOOL      bSuccessFork = FALSE;
	pid_t     pid;

	assert(s_pEngine);
	assert(index >= 0);

	/* Try to fork a process */
	while (!bSuccessFork)
	{
		switch ((pid = fork()))
		{
		case -1:
			hserrlog("fork");
			sleep(1);
			continue;
		case 0:
			bSuccessFork = TRUE;
			break;
		default:
			hslog(2, "worker forked [pid:%d]", pid);
			close(sockWorker);
			return TRUE;
		}
	}
	/* end fork; child process will go on */

}

static BOOL do_service_forked()
{
	SOCKET    sockWorker = (SOCKET) - 1;
	hs_handler *pfHandler;
	hs_ctx_t  ctxWorker;

	int       n;
	hslist_t  i;

	assert(s_pEngine);

	n = s_pEngine->nInterestings;

	if (n <= 0)
	{
		return FALSE;
	}

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (n <= 0)
		{
			break;
		}
		if (!HSCTX(i)->interesting)
		{
			continue;
		}

		/* Selectable interface */
		{
			struct sockaddr_in addr;
			int       size = sizeof addr;

			ctxWorker.sock = accept(HSCTX(i)->sock, (struct sockaddr *) &addr, &size);
			ctxWorker.state = HSS_INIT;
			break;
		}
		n--;
		HSCTX(i)->interesting = FALSE;
	}

	assert(!IS_HSLIST_END(i) && n > 0);

	pfHandler = HSCTX(i)->handler;

	assert(ctxWorker.sock != (SOCKET) - 1);
	assert(pfHandler);

	(*pfHandler) (&ctxWorker);
	sleep(0);
	shutdown(ctxWorker.sock, 2);
	hslog(2, "wait peer close");
	select_a_socket(ctxWorker.sock, 10);
	close(ctxWorker.sock);
}

void hs_run_forked()
{
	while (!s_pEngine->bTerminated)
	{
		if (!select_sockets(s_pEngine))
		{
			return;
		}
		do_service_forked(s_pEngine);
	}
}

static void keep_processes()
{
	while (!s_pEngine->bTerminated && s_pEngine->nCurrProcesses < s_pEngine->nMinProcesses)
	{
		pid_t     pid;
		switch ((pid = fork()))
		{
		case -1:
			hserrlog("fork");
			sleep(1);
			continue;
		case 0:
			hs_run_forked(s_pEngine);
			exit(0);
		default:
			s_pEngine->nCurrProcesses++;
			hslog(2, "forked [pid:%d]", pid);
		}
	}
}

void hs_run()
{
	if (FALSE == create_sockets(s_pEngine))
	{
		return;
	}
	if (FALSE == listen_sockets(s_pEngine))
	{
		return;
	}

	if (s_pEngine->nSockets > 1)
	{
		hs_lock_init();
	}
	while (!s_pEngine->bTerminated)
	{
		keep_processes(s_pEngine);
		sleep(300);
	}
}

/*------------- USER ---------------*/

int webserver_core(int fd, void *arg)
{
	const char *msg = "200 OK\r\nContent-type: text/plain\r\n\r\nTest page\r\n";
	char      req[LARGE_BUFFER_SIZE];
	if (0 < recv(fd, req, sizeof req, 0))
	{
		send(fd, msg, strlen(msg), 0);
	}
	return 0;
}

hs_state_t webserver(hs_ctx_t * ctx)
{
	int       fd = ctx->sock;
	webserver_core(fd, 0);
	return 0;
}

int main()
{
	hs_init(HS_PREFORKED);
	hs_cntl(HS_PREFORKS, 5);
	hs_server_add("0.0.0.0", 8080, webserver);
	hs_server_add("0.0.0.0", 8088, webserver);
	hs_run();
	hs_destroy();
}

/*
 * vim: set path+=../src
 */
