/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: model4.c 101 2004-07-19 15:45:30Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief Model 4 thinkpad..
 *  model4 is one process non-block I/O type.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"
#include "hslog.h"
#include "hslist.h"

enum _HS_STATE
{
	HSS_STATE_BEGIN = 1000,
	HSS_INIT,
	HSS_ACCEPT,
	HSS_ONACCEPT,
	HSS_RECV,
	HSS_ONRECV,
	HSS_SEND,
	HSS_ONSENT,
	HSS_CLOSE,
	HSS_ONCLOSE,
	HSS_ONERROR,
	HSS_STATE_END
};
typedef enum _HS_STATE hs_state_t;

typedef struct _hs_ctx_t hsocket_t;
typedef hs_state_t hs_handler(hsocket_t *);

struct _hs_ctx_t
{
	hs_state_t state;				   /* CURRENT STATE */
	int       fd;					   /* SOCKET */
	BOOL      interesting;				   /* SELECTED FOR ONACCEPT, ONRECV, ONCLOSE */
	int       errnum;				   /* ONERROR */
	struct sockaddr_in address;			   /* SERVER SOCKET BIND ADDRESS */

	          hs_state_t(*handler) (hsocket_t *);	   /* SOCKET HANDLER */
	int       sendwaits;
	char     *sendbuffer;
	void     *argp;					   /* FOR USER USAGE */
	long      argl;					   /* FOR USER USAGE */
};

typedef struct _hserver_engine
{
	int       nType;
	int       nInterestings;
	int       nMaxProcesses;
	int       nMinProcesses;
	int       nCurrProcesses;
	int       nSockets;
	hslist_t  lstSockets;
	BOOL      bTerminated;
}
hserver_engine_t;

#define HSCTX(i) ((hsocket_t*)((i)->data))

enum _hs_type
{
	HS_FORK,
	HS_THREAD,
	HS_PREFORKED,
	HS_NONBLOCKIO,
	HS_PREFORKED_NONBLOCKIO,
	HS_THREAD_NONBLOCKIO
};

enum _hs_cntl_type
{
	HS_PREFORKS,
	HS_PREFORK_MIN,
	HS_PREFORK_MAX
};

/* LOCAL DATA -----------------------------------------------------------------------*/

static hserver_engine_t s_engine;
static hserver_engine_t *s_pEngine = &s_engine;

/* FUNCTION DEFINITION --------------------------------------------------------------*/

hsocket_t *hs_ctx_new(void)
{
	return (hsocket_t *) malloc(sizeof(struct _hs_ctx_t));
}

BOOL hs_cntl(int nType, int nValue)
{
	assert(s_pEngine);

	switch (nType)
	{
	case HS_PREFORKS:
		s_pEngine->nMaxProcesses = nValue;
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MIN:
		s_pEngine->nMinProcesses = nValue;
		break;
	case HS_PREFORK_MAX:
		s_pEngine->nMaxProcesses = nValue;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

static void hs_signal(int sig, siginfo_t * sf, void *vp)
{
	pid_t     pid;
	int       status;

	switch (sig)
	{
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		{
			hslog(2, "worker terminated [pid:%d]", pid);
			s_pEngine->nCurrProcesses--;

			/* none */
		}
		break;
	case SIGTERM:
	case SIGINT:
		s_pEngine->bTerminated = TRUE;
		hs_cntl(HS_PREFORKS, 0);
		hslog(2, "Terminating..");
		break;
	default:
		hserrlog("Unknown signal caught [%d:%s]", sig, strsignal(sig));
		assert(0 && "Can't handle this signal");
		break;
	}
}

BOOL hs_init(int nType)
{
	struct sigaction sa;

	assert(s_pEngine);

	hslog(2, "hserver engine init");
	memset(s_pEngine, 0, sizeof *s_pEngine);

	s_pEngine->nType = nType;
	s_pEngine->bTerminated = FALSE;
	hslist_init(&s_pEngine->lstSockets);

	memset(&sa, 0, sizeof sa);
	sa.sa_sigaction = &hs_signal;
	hslog(2, "SIGCHLD/SIGTERM/SIGINT interrupt handler registered");
	sigaction(SIGCHLD, &sa, 0);
	sigaction(SIGTERM, &sa, 0);
	sigaction(SIGINT, &sa, 0);
	return TRUE;
}

void hs_destroy()
{
	hslist_t  i;

	hslog(2, "hserver shutdown");
	if (!s_pEngine->nSockets)
	{
		return;
	}
	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (HSCTX(i)->fd != (SOCKET) - 1)
		{
			if (close(HSCTX(i)->fd))
			{
				hslog(1, "server socket [%d] closing error", HSCTX(i)->fd);
			}
		}
	}
}

void hs_server_add(const char *ip, int nPort, hs_handler * pfHandler)
{
	hsocket_t *sock;

	assert(pfHandler);

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	hslog(2, "server socket added [%s:%d]", ip, nPort);

	sock = hs_ctx_new();
	if (!sock)
	{
		TODO("Memory allocation failure.");
		return;
	}
	sock->address.sin_family = AF_INET;
	sock->address.sin_port = htons(nPort);
	sock->handler = pfHandler;
	hslist_push_front(&s_pEngine->lstSockets, sock);
	s_pEngine->nSockets++;
}

hsocket_t *hsocket_server(const char *ip, int port)
{
	hsocket_t *sock;
	int       reuse = 1;
	struct sockaddr_in addr;

	if (!ip || !*ip)
	{
		ip = "0.0.0.0";
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	inet_aton(ip, &(addr.sin_addr));

	sock = malloc(sizeof(*sock));
	if (!sock)
	{
		return 0;
	}

	sock->fd = socket(PF_INET, SOCK_STREAM, 0);
	if (sock->fd < 0)
	{
		hserrlog("socket");
	}
	else if (0 > setsockopt(sock->fd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
	}
	else if (0 == bind(sock->fd, (struct sockaddr *) &addr, sizeof addr))
	{
		return sock;
	}
	hserrlog("bind");
	free(sock);
	return 0;
}

static SOCKET create_a_socket(const struct sockaddr_in *pAddr)
{
	SOCKET    s;
	int       reuse = 1;

	assert(pAddr);

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		hserrlog("socket");
		return s;
	}
	{
		int       val;
		val = fcntl(s, F_GETFL, 0);
		fcntl(s, F_SETFL, val | O_NONBLOCK);
	}
#if 0
	{
		val = ioctl(s, FIONBIO, 1);
	}
#endif
	if (0 > setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse))
	{
		hserrlog("setsockopt");
		return -1;
	}
	if (0 > bind(s, (struct sockaddr *) pAddr, sizeof *pAddr))
	{
		hserrlog("bind");
		return -1;
	}
	return s;
}

static BOOL create_sockets()
{
	hslist_t  i;

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (0 > (HSCTX(i)->fd = create_a_socket(&(HSCTX(i)->address))))
		{
			return FALSE;
		}
	}
	return TRUE;
}

static BOOL listen_sockets()
{
	hslist_t  i;

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		assert(HSCTX(i)->fd >= 0);
		if (0 > listen(HSCTX(i)->fd, 5))
		{
			return FALSE;
		}
		hslog(2, "listen %d", HSCTX(i)->fd);
	}
	return TRUE;
}

static BOOL select_sockets()
{
	hslist_t  i;
	int       max = -1;
	int       nCount = 0;
	fd_set    fdRead;
	fd_set    fdWrite;

	assert(s_pEngine);
      again:
	FD_ZERO(&fdRead);
	FD_ZERO(&fdWrite);

	hslog(1, "select sockets: %d", s_pEngine->nSockets);

	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		SOCKET    s = HSCTX(i)->fd;
		assert(s);
		FD_SET(s, &fdRead);
		if (HSCTX(i)->nSendWaits)
		{
			FD_SET(s, &fdWrite);
		}
		if (max < s)
		{
			max = s;
		}
		HSCTX(i)->interesting = FALSE;
	}

	if (0 >= (nCount = select(max + 1, &fdRead, 0, 0, 0)))
	{
		if (errno == EINTR && !s_pEngine->bTerminated)
		{
			goto again;
		}
		return FALSE;
	}
	hslog(2, "selected socket count [%d]", nCount);

	s_pEngine->nInterestings = nCount;
	HSLIST_FOR_EACH(i, s_pEngine->lstSockets)
	{
		if (FD_ISSET(HSCTX(i)->fd, &fdRead))
		{
			HSCTX(i)->interesting = TRUE;
			nCount--;
		}
	}

	return TRUE;
}

static BOOL select_a_socket(SOCKET fd, int timeout)
{
	fd_set    fdRead, fdWrite, fdExcpt;
	struct timeval tval;
	int       ret;

	FD_ZERO(&fdRead);
	FD_ZERO(&fdWrite);
	FD_ZERO(&fdExcpt);

	FD_SET(fd, &fdRead);
	FD_SET(fd, &fdWrite);
	FD_SET(fd, &fdExcpt);

	if (timeout >= 0)
	{
		tval.tv_sec = timeout;
		tval.tv_usec = 0;
		ret = select(fd + 1, &fdRead, &fdWrite, &fdExcpt, &tval);
	}
	else
	{
		ret = select(fd + 1, &fdRead, &fdWrite, &fdExcpt, 0);
	}
	return ret == 1;
}

void hs_run()
{
	if (FALSE == create_sockets(s_pEngine))
	{
		return;
	}
	if (FALSE == listen_sockets(s_pEngine))
	{
		return;
	}
	while (!s_pEngine->bTerminated)
	{
		select_sockets();
		sleep(300);
	}
}

/*------------- USER ---------------*/

hs_state_t webserver_recv(int fd, void *arg)
{
	const char *msg = "200 OK\r\nContent-type: text/plain\r\n\r\nTest page\r\n";
	char      req[LARGE_BUFFER_SIZE];
	if (0 < recv(fd, req, sizeof req, 0))
	{
		return HSS_CLOSE;
	}
	return HSS_RECV;
	send(fd, msg, strlen(msg), 0);
	return 0;
}

hs_state_t webserver(hsocket_t * sock, int state)
{
	int       fd = sock->fd;
	switch (state)
	{
	case HSS_RECV:
		return webserver_recv(fd, 0);
	case HSS_SENT:
		return webserver_sent(fd, 0);
	default:
		assert(0);
	}
	return HSS_CLOSE;
}

int main()
{
	hs_init(HS_NONBLOCKIO);
	hs_server_add("0.0.0.0", 8080, webserver);
	hs_server_add("0.0.0.0", 8088, webserver);
	hs_run();
	hs_destroy();
	return 0;
}

/*
 * vim: set path+=../src
 */
