/*
 * This is a part of hserver project.                                                
 * You can freely redistribute this source code under BSD license
 */
/* $Id: test_client.c 100 2004-07-19 15:45:07Z pynoos $ */

/* FILE DESCRIPTION -----------------------------------------------------------------*/

/*! \author Hojin Choi
 *  \brief Connect storm
 * This program generates connect storm and send simple query.
 */

/* FORWARD DECLARATION --------------------------------------------------------------*/
#include "hserverstd.h"

struct sock_state
{
	int       state;
	int       sock;
	int       select_req;
	int       interesting;
};

enum SOCKET_STATE
{
	SS_BEGIN,
	SS_CREATED,
	SS_CONNECTING,
	SS_CONNECTED,
	SS_SENT,
	SS_RECEIVING,
	SS_RECEIVED,
	SS_ERROR,
	SS_END
};

#define PRINTF if(s_debug) printf
static int s_debug = 0;
static int s_count_try;
static int s_count_connected;
static int s_count_error;
static time_t s_time_begin;
static time_t s_timeout;
static char request_path[2048];
static int request_path_len;

/* LOCAL DATA -----------------------------------------------------------------------*/

/* FUNCTION DEFINITION --------------------------------------------------------------*/

/*! \brief This is sample function
 *
 * This sample function is from someone's brains.
 * It works very well in such a complicated environment even with
 * no line in those body.
 *
 * \param arg1 result will be stored in this non-const pointing buffer
 * \param arg2 the arg1's buffer length
 * \param arg3 arg3 is non-editable string!
 * \return 0 : command was done and no error was found.
 *         -1: command was failed to cry..
 * \see See also
 */

static void init_state(struct sock_state *pSockState)
{
	pSockState->select_req = 0;
	pSockState->interesting = 1;
	pSockState->sock = socket(PF_INET, SOCK_STREAM, 0);
	if (pSockState->sock >= 0)
	{
		if (0 > fcntl(pSockState->sock, F_SETFL, O_NONBLOCK))
		{
			perror("fcntl");
			TODO("We need error control..");
		}
	}
	PRINTF("%d: created\n", pSockState->sock);
	pSockState->state = SS_CREATED;
}

static void send_message(struct sock_state *pSockState)
{
	int       s;
	PRINTF("%d: sending..\n", pSockState->sock);
	s = send(pSockState->sock, request_path, request_path_len, 0);
	if (s < 0)
	{
		perror("send");
		pSockState->state = SS_ERROR;
		pSockState->select_req = 0;
		pSockState->interesting = 1;
		return;
	}
	pSockState->state = SS_SENT;
	pSockState->select_req = 1;
	pSockState->interesting = 0;
}

static void recv_message(struct sock_state *pSockState)
{
	char      buf[128];
	int       r;
	PRINTF("%d: receiving..\n", pSockState->sock);
	r = recv(pSockState->sock, buf, sizeof buf, 0);
	pSockState->select_req = 0;
	pSockState->interesting = 1;
	if (r < 0)
	{
		perror("recv");
		pSockState->state = SS_ERROR;
		return;
	}
	buf[r] = 0;
	pSockState->state = SS_END;
	return;
}

static void connect_target(const struct sockaddr_in *addr, struct sock_state *pSockState)
{
	socklen_t len = sizeof *addr;

	if (0 != connect(pSockState->sock, (const struct sockaddr *) addr, len))
	{
		if (errno != EINPROGRESS)
		{
			pSockState->state = SS_ERROR;
			perror("connect");
			return;
		}
	}
	PRINTF("%d: connecting..\n", pSockState->sock);
	pSockState->state = SS_CONNECTING;
	pSockState->select_req = 1;
	pSockState->interesting = 0;
	s_count_try++;
	return;
}

static int get_error(struct sock_state *pSockState)
{
	int       val;
	int       len = sizeof val;
	if (0 > getsockopt(pSockState->sock, SOL_SOCKET, SO_ERROR, &val, &len))
	{
		perror("getsockopt");
		return 0;
	}
	return val;
}

static int prepare_sockets(struct sock_state *pSockStates, int max)
{
	int       i = 0;
	for (i = 0; i < max; i++)
	{
		pSockStates[i].state = SS_BEGIN;
		pSockStates[i].sock = -1;
		pSockStates[i].select_req = 0;
		pSockStates[i].interesting = 1;
	}
	return i;
}

static void prepare_address(const char *addr, int port, struct sockaddr_in *pAddr)
{
	pAddr->sin_family = AF_INET;
	pAddr->sin_port = htons(port);
	inet_aton(addr, &(pAddr->sin_addr));
}

void wait_event(struct sock_state *pSockStates, int max)
{
	fd_set    fdr, fdw;
	int       i;
	int       maxfd = -1;
	int       count;

	FD_ZERO(&fdr);
	FD_ZERO(&fdw);
	for (i = 0; i < max; i++)
	{
		if (!pSockStates[i].select_req)
		{
			continue;
		}
		if (maxfd < pSockStates[i].sock)
		{
			maxfd = pSockStates[i].sock;
		}
		if (pSockStates[i].state == SS_CONNECTING
		    || pSockStates[i].state == SS_CONNECTED || pSockStates[i].state == SS_SENT)
		{
			FD_SET(pSockStates[i].sock, &fdw);
		}
		else if (pSockStates[i].state == SS_RECEIVING)
		{
			FD_SET(pSockStates[i].sock, &fdr);
		}
	}
	if (-1 == maxfd)
	{
		return;
	}
	PRINTF("Waiting..\n");
      again:
	count = select(maxfd + 1, &fdr, &fdw, 0, 0);
	if (0 > count)
	{
		if (errno == EINTR)
		{
			goto again;
		}
		perror("select");
		return;
	}
	for (i = 0; i < max && count > 0; i++)
	{
		if (pSockStates[i].select_req && FD_ISSET(pSockStates[i].sock, &fdr))
		{
			count--;
			pSockStates[i].interesting = 1;
		}
		if (pSockStates[i].select_req && FD_ISSET(pSockStates[i].sock, &fdw))
		{
			count--;
			pSockStates[i].interesting = 1;
		}
	}
}

static int report()
{
	int       elapsed;
	int       try_per_sec, connected_per_sec;
	elapsed = time(0) - s_time_begin;

	assert(elapsed >= 0);
	if (0 == elapsed)
	{
		return 0;
	}

	try_per_sec = (s_count_try * 10 / elapsed);
	connected_per_sec = (s_count_connected * 10 / elapsed);
	printf("Elapsed       : %d sec.\n", elapsed);
	printf("Try           : %d\n", s_count_try);
	printf("Connected     : %d\n", s_count_connected);
	printf("Error         : %d\n", s_count_error);
	printf("Try/sec       : %d.%d/sec.\n", try_per_sec / 10, try_per_sec % 10);
	printf("Connected/sec : %d.%d/sec.\n", connected_per_sec / 10, connected_per_sec % 10);
	printf("\n");
	return 0;
}

static int run_sockets(const struct sockaddr_in *targetaddr, struct sock_state *pSockStates, int max)
{
	int       i = 0;

	assert(targetaddr);
	assert(pSockStates);
	assert(max > 0);

	s_time_begin = time(0);
	alarm(1);

	while (1)
	{
		for (i = 0; i < max; i++)
		{
			PRINTF("%d: State: %d Interesting: %d SelectReq: %d\n",
			       pSockStates[i].sock,
			       pSockStates[i].state, pSockStates[i].interesting, pSockStates[i].select_req);
			if (!pSockStates[i].interesting)
			{
				continue;
			}

			assert(pSockStates[i].state >= 0 && pSockStates[i].state <= SS_END);

			switch (pSockStates[i].state)
			{
			case SS_BEGIN:
				init_state(&pSockStates[i]);
				break;
			case SS_CREATED:
				connect_target(targetaddr, &pSockStates[i]);
				break;
			case SS_CONNECTING:
				if (0 != get_error(&pSockStates[i]))
				{
					pSockStates[i].state = SS_ERROR;
					s_count_error++;
				}
				else
				{
					pSockStates[i].state = SS_CONNECTED;
					s_count_connected++;
				}
				pSockStates[i].select_req = 0;
				pSockStates[i].interesting = 1;
				break;
			case SS_CONNECTED:
				send_message(&pSockStates[i]);
				break;
			case SS_SENT:
				PRINTF("%d: sent\n", pSockStates[i].sock);
				pSockStates[i].state = SS_RECEIVING;
				pSockStates[i].select_req = 1;
				pSockStates[i].interesting = 0;
				break;
			case SS_RECEIVING:
				recv_message(&pSockStates[i]);
				break;
			case SS_ERROR:
			case SS_END:
				PRINTF("%d: close\n", pSockStates[i].sock);
				close(pSockStates[i].sock);
				pSockStates[i].state = SS_BEGIN;
				pSockStates[i].select_req = 0;
				pSockStates[i].interesting = 1;
				break;
			}
		}
		wait_event(pSockStates, max);
		sleep(0);
	}
	return i;
}

static void help()
{
	fprintf(stderr, "Usage: test_client <ip> <port> <seconds> [<request path>]\n"
		"       connect to <ip>:<port> and send HTTP Request as below:\n\n"
		"       GET /\\r\\n\\r\\n\n\n"
		"       or\n\n"
		"       GET <request path>\\r\\n\\r\\n\n\n"
		"       You can send connect-storm to that address as a form of HTTP during given seconds\n"
		"       Made by Hojin Choi <pynoos@users.kldp.net>\n" "       Test program of hserver project\n");
}

static void timeout(int sig)
{
	report();
	s_timeout--;
	if (s_timeout <= 0)
	{
		exit(0);
	}
	alarm(1);
}

int main(int argc, char *argv[])
{
	struct sockaddr_in target_addr;
	struct sock_state ss[512];

	if (argc < 4)
	{
		help();
		exit(1);
	}
	assert(argc >= 4);

	strcpy(request_path, "GET ");
	if (argc > 4)
	{
		strncat(request_path, argv[4], sizeof request_path - strlen(request_path));
	}
	else
	{
		strcat(request_path, "/");
	}
	strncat(request_path, " HTTP/1.0\r\n\r\n", sizeof request_path - strlen(request_path));
	request_path_len = strlen(request_path);

	signal(SIGALRM, timeout);
	s_timeout = atoi(argv[3]);

	prepare_sockets(ss, sizeof ss / sizeof *ss);
	prepare_address(argv[1], atoi(argv[2]), &target_addr);

	run_sockets(&target_addr, ss, sizeof ss / sizeof *ss);
	return 0;
}

#ifndef NDEBUG
static char id[] __UNUSED__ = "$Id: test_client.c 100 2004-07-19 15:45:07Z pynoos $";
#endif
/*
 * History
   $Log$
   Revision 1.5  2004/07/19 15:45:07  pynoos
   Add request path

   Revision 1.4  2003/10/07 08:53:55  pynoos
   Fix not to print trace log

   Revision 1.3  2003/10/07 07:20:49  pynoos
   Test

   Revision 1.2  2003/09/08 17:28:22  pynoos
   Model1 completed.

   Revision 1.1  2003/08/31 17:12:01  pynoos
   New added test program

 */

/*
 * vim: path+=../src
 */
